import re


class ImportResult:
    def __init__(self):
        self.added = {}
        self.updated = {}
        self.unchanged = {}
        self.surplus = {}
        self.failed = {}
        self.errors = []

    def set_surplus(self, record):
        id = record["id"]
        self.surplus[id] = record

    def set_added(self, record):
        id = record["id"]
        self.added[id] = record
        self.surplus.pop(id)

    def set_updated(self, record):
        id = record["id"]
        self.updated[id] = record
        self.surplus.pop(id)

    def set_unchanged(self, record):
        id = record["id"]
        self.unchanged[id] = record
        self.surplus.pop(id)

    def set_deleted(self, record):
        self.surplus.pop(record["id"])

    def fail(self, record, error):
        id = record["id"]
        self.failed[id] = record
        self.errors.append(error)

    def dict(self):
        return dict(
            added=self.added,
            updated=self.updated,
            unchanged=self.unchanged,
            surplus=self.surplus,
            failed=self.failed,
            errors=self.errors,
        )


class Importer:
    def __init__(self, api):
        self.api = api
        self.api.raw = False
        self.api.json = True
        self.api.by_id = True
        self.api.include_id = True
        self.api.include_key = True
        self.records_by_id = {}
        self.records_by_key = {}
        self.results = {}

    def import_records(self, records):
        if not isinstance(records, list):
            if self.get_field(records, "type", require=False, allow_none=True) in self.api.IMPORT_TYPES:
                # assume input is a single record
                records = [records]
            else:
                # assume input is a dict keyed bye ID or Key
                records = records.values()

        for record in records:
            zone = self.get_field(record, "zone")
            self.get_zone_records(zone)
            try:
                self.import_record(zone, record)
            except Exception as e:
                self.results[zone].fail(record, str(e))

        return {zone: result.dict() for zone, result in self.results.items()}

    def get_zone_records(self, zone):
        if zone not in self.records_by_id:
            self.records_by_id[zone] = {}
            self.records_by_key[zone] = {}
            records = self.api.get_zone_records(zone)
            formatted = self.api.format_records(zone, records)
            self.results[zone] = ImportResult()
            for id, record in formatted.items():
                key = self.get_field(record, "key")
                self.records_by_id[zone][id] = record
                self.records_by_key[zone][key] = record
                self.results[zone].set_surplus(record)

        return self.records_by_id[zone].values()

    def get_record_by_id(self, zone, id):
        return self.records_by_id[zone].get(id, None)

    def get_record_by_key(self, zone, key):
        return self.records_by_key[zone].get(key, None)

    def get_field(self, record, field, *, require=True, allow_none=False, default=None):
        if field in record:
            value = record.get(field, default)
            if value is None:
                if not allow_none:
                    raise RuntimeError(f"Null '{field}' value in input record")
        else:
            if require:
                raise RuntimeError(f"Missing field '{field}' in input record")
            value = default
        return value

    def import_record(self, zone, record):
        current = None

        if "id" in record:
            id = self.get_field(record, "id")
            current = self.get_record_by_id(zone, id)

        if current is None:
            if "key" in record:
                key = self.get_field(record, "key")
            else:
                key = self.api.format_key(zone, record)
            current = self.get_record_by_key(zone, key)

        if current is None:
            current = self.find_matching_current_record(zone, record)
            if current is not None:
                self.results[zone].set_unchanged(current)
                return

        updating = False
        if current is not None:
            id = self.get_field(current, "id")
            if self.diff(record, current):
                updating = True
                self.delete_record(zone, id)
                self.results[zone].set_deleted(current)
            else:
                self.results[zone].set_unchanged(current)
                return

        added = self.add_record(zone, record)
        if updating:
            self.results[zone].set_updated(added)
        else:
            self.results[zone].set_added(added)
        return

    def delete_record(self, zone, id):
        record = dict(id=id)
        return self.api.delete_record(zone, record)

    def find_matching_current_record(self, zone, record):
        """find matching current record in zone records"""
        current_records = self.get_zone_records(zone)
        for current in current_records:
            if self.get_field(current, "type") in self.api.IMPORT_TYPES:
                if not self.diff(record, current):
                    return current
        return None

    def diff(self, record, current):
        """check record content for difference"""
        type = self.get_field(record, "type")
        if type not in self.api.IMPORT_TYPES:
            raise RuntimeError(f"cannot import type '{type}'")
        if type != self.get_field(current, "type"):
            return True
        if self.get_field(record, "zone") != self.get_field(current, "zone"):
            raise RuntimeError("zone mismatch")
        if self.get_field(record, "host") != self.get_field(current, "host"):
            return True
        if self.get_field(record, "content") != self.get_field(current, "content"):
            return True
        if self.get_field(record, "ttl", require=False, default=self.api.DEFAULT_TTL) != self.get_field(current, "ttl"):
            return True
        if type in self.api.PRIORITY_TYPES:
            if self.get_field(record, "priority", require=False, default=self.api.DEFAULT_PRIORITY) != self.get_field(
                current, "priority"
            ):
                return True
        if type == "SRV":
            if self.get_field(record, "weight", require=False, default=self.api.DEFAULT_WEIGHT) != self.get_field(
                current, "weight"
            ):
                return True
            if self.get_field(record, "port") != self.get_field(current, "port"):
                return True
        return False

    def add_record(self, zone, record):
        type = self.get_field(record, "type")
        if type not in self.api.IMPORT_TYPES:
            raise ValueError(f"Unsupported record type '{type}'")
        host = self.get_field(record, "host")
        content = self.get_field(record, "content")
        ttl = self.get_field(record, "ttl", require=False, default=self.api.DEFAULT_TTL)
        priority = None
        weight = None
        port = None
        if type in self.api.PRIORITY_TYPES:
            priority = self.get_field(record, "priority", require=False, default=self.api.DEFAULT_PRIORITY)
        if type == "SRV":
            matched = re.match(r"^(\d+) (\d+) (.*)$", content)
            if matched:
                weight, port, content = matched.groups()
                if "weight" in record and int(weight) != int(self.get_field(record, "weight")):
                    raise ValueError("content weight != record field")
                if "port" in record and int(port) != int(self.get_field(record, "port")):
                    raise ValueError("content port != record field")
            else:
                weight = self.get_field(record, "weight")
                port = self.get_field(record, "port")
            if port is None:
                raise ValueError("port is required for SRV record")
        added = self.api.add_record(zone, type, host, content, ttl, priority, weight, port)
        return self.api.format_record(zone, added)



"""Cloudflare API Client"""

import json
import os
import re

import CloudFlare


class API:

    RECORD_TYPES = ["A", "AAAA", "CNAME", "MX", "NS", "SOA", "TXT", "SRV", "LOC", "SSHFP", "CAA", "TLSA"]
    IMPORT_TYPES = ["A", "AAAA", "CNAME", "MX", "TXT", "SRV"]
    PRIORITY_TYPES = ["MX", "SRV"]

    DEFAULT_PRIORITY = 10
    DEFAULT_WEIGHT = 5
    DEFAULT_TTL = 60

    MAX_ZONES = 128

    def __init__(
        self,
        *,
        token=None,
        json=False,
        quiet=False,
        raw=False,
        include_id=False,
        include_ttl=False,
        include_key=False,
        output_function=None,
        by_key=False,
        by_id=False,
    ):
        token = token or os.environ["CLOUDFLARE_AUTH_TOKEN"]
        self.client = CloudFlare.CloudFlare(token=token)
        self.json = json
        self.raw = raw
        self.quiet = quiet
        self.include_id = include_id
        self.include_ttl = include_ttl
        self.include_key = include_key
        self.by_key = by_key
        self.by_id = by_id
        self.output_function = output_function or print

    def output(self, data):
        if self.quiet:
            return
        if self.json:
            data = json.dumps(data, indent=2)
        elif not data:
            return
        self.output_function(data)

    def get_zones(self):
        return self.client.zones.get(params={"per_page": self.MAX_ZONES})

    def get_zone_id(self, domain):
        zones = self.get_zones()
        for z in zones:
            if z["name"] == domain:
                return z["id"]
        raise ValueError(f"unknown domain {domain}")

    def get_zone_records(self, domain):
        records = {}
        zone_id = self.get_zone_id(domain)
        records = self.client.zones.dns_records.get(zone_id, params={"per_page": self.MAX_ZONES})
        return records

    def parse_host(self, name, domain):
        if name in ["@", domain]:
            name = domain
        else:
            name = ".".join([name, domain])
        name = name.strip(".")
        return name

    def delete_records(self, domain, records):
        deleted = [self.delete_record(domain, record) for record in records]
        if self.json:
            return deleted
        else:
            return "\n".join(deleted)

    def delete_record(self, domain, record):
        record_id = record["id"]
        zone_id = self.get_zone_id(domain)
        ret = self.client.zones.dns_records.delete(zone_id, record_id)
        if self.json:
            return ret
        else:
            return ret["id"]

    def add_record(self, domain, type, host, content, ttl=None, priority=None, weight=None, port=None):
        zone_id = self.get_zone_id(domain)
        host = self.parse_host(host, domain)
        if ttl is None:
            ttl = self.DEFAULT_TTL
        if type in self.PRIORITY_TYPES and priority is None:
            priority = self.DEFAULT_PRORITY
        record = dict(type=type, name=host, content=content, ttl=ttl)
        if type == "MX":
            record["priority"] = priority
        elif type == "SRV":
            if port is None:
                raise ValueError("port is required for SRV record")
            if weight is None:
                weight = self.DEFAULT_WEIGHT
            record["data"] = dict(port=port, priority=priority, target=record.pop("content"), weight=weight)
        ret = self.client.zones.dns_records.post(zone_id, data=record)
        if self.json:
            return ret
        else:
            return ret["id"]

    def update_records(self, domain, records):
        updated = [self.update_record(domain, record) for record in records]
        if self.json:
            return updated
        else:
            return "\n".join(updated)

    def update_record(self, domain, record):
        dns_record_id = record["id"]
        zone_id = self.get_zone_id(domain)
        update = dict(
            name=record["name"],
            type=record["type"],
            content=record["content"],
            ttl=record["ttl"],
            proxied=record["proxied"],
        )
        if record["type"] == "SRV":
            # SRV record update uses data dict for port, priority, weight, target
            update.pop("content")
            update["data"] = record["data"]
        elif record["type"] == "MX":
            update["priority"] = record["priority"]
        ret = self.client.zones.dns_records.patch(zone_id, dns_record_id, data=update)

        if self.json:
            return ret
        else:
            return ret["id"]

    def format_host(self, domain, record):
        if "host" in record:
            return record["host"]
        host = record["name"]
        if host.endswith(domain):
            host = host[: -1 - len(domain)]
        if not host:
            host = "@"
        return host

    def format_record(self, domain, record):
        if self.raw:
            return record

        out = dict(
            id=record["id"],
            key=self.format_key(domain, record),
            zone=domain,
            host=self.format_host(domain, record),
            fqdn=record["name"],
            type=record["type"],
            content=record["content"],
            priority=record.get("priority", record.get("data", {}).get("priority", None)),
            weight=record.get("data", {}).get("weight", None),
            port=record.get("data", {}).get("port", None),
            protocol=self.format_protocol(domain, record),
            ttl=record["ttl"],
        )

        if self.json:
            return out

        fields = []
        if self.include_id:
            fields.append(out["id"])
        if self.include_key:
            fields.append(out["key"])
        fields.append(out["type"])
        fields.append(out["protocol"])
        fields.append(out["priority"])
        fields.append(out["host"])
        if record["type"] == "TXT":
            fields.append('"' + out["content"] + '"')
        else:
            fields.append(out["content"])
        if self.include_ttl:
            fields.append(out["ttl"])
        return " ".join([str(f) for f in fields if f not in ["", None, "null", "None", False]])

    def format_records(self, domain, records):
        formatted = [self.format_record(domain, r) for r in records]
        if self.raw:
            return formatted
        if self.json:
            if self.by_key:
                return {f["key"]: f for f in formatted}
            elif self.by_id:
                return {f["id"]: f for f in formatted}
            return formatted
        else:
            return "\n".join(formatted)

    def is_selected(self, pattern, text):
        if pattern is None:
            return True
        elif pattern.startswith("/"):
            return bool(re.match(pattern.strip("/"), text))
        else:
            return pattern == text

    def select_records(self, domain, type=None, host=None, content=None, priority=None, weight=None):

        if host and type != "ID" and not host.startswith("/"):
            host = self.parse_host(host, domain)

        records = self.get_zone_records(domain)

        selected = []

        if type == "ID":
            for record in records:
                if host == record["id"]:
                    return [record]
            return []

        for record in records:
            if not self.is_selected(type, record["type"]):
                continue
            if not self.is_selected(host, record["name"]):
                continue
            if not self.is_selected(content, record["content"]):
                continue
            if record["type"] in ["MX", "SRV"]:
                if priority is not None:
                    if int(priority) != int(record["priority"]):
                        continue
            if record["type"] == "SRV":
                if weight is not None:
                    if int(weight) != int(record["weight"]):
                        continue

            selected.append(record)

        return selected

    def format_key(self, domain, record):
        fields = []
        if record["type"] == "TXT":
            fields.append(record["type"] + "." + self.format_protocol(domain, record))
        elif record["type"] in ["MX", "SRV"]:
            fields.append(record["type"] + "." + str(record.get("priority", self.DEFAULT_PRIORITY)))
        else:
            fields.append(record["type"])
        fields.append(self.format_host(domain, record))
        fields.append(domain)
        return ";".join(fields)

    def format_protocol(self, domain, record):
        if record["type"] != "TXT":
            return None
        if "name" in record:
            fqdn = record["name"]
        else:
            fqdn = record["fqdn"]
        if fqdn.endswith(f"._domainkey.{domain}"):
            return "DKIM"
        content = record["content"]
        patterns = {
            "^v=([a-zA-Z]+).*": None,
            "^(mailconf)=.*": None,
            "^k=rsa;.*": "DKIM",
        }
        for pattern, proto in patterns.items():
            m = re.match(pattern, content)
            if m:
                if proto:
                    return proto.upper()
                if len(m.groups()) > 0:
                    return m.groups()[0].upper()
        return None
